/*
<:copyright-gpl 
 Copyright 2002 Broadcom Corp. All Rights Reserved. 
 
 This program is free software; you can distribute it and/or modify it 
 under the terms of the GNU General Public License (Version 2) as 
 published by the Free Software Foundation. 
 
 This program is distributed in the hope it will be useful, but WITHOUT 
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
 for more details. 
 
 You should have received a copy of the GNU General Public License along 
 with this program; if not, write to the Free Software Foundation, Inc., 
 59 Temple Place - Suite 330, Boston MA 02111-1307, USA. 
:>
*/
/***************************************************************************
 * File Name  : bcm63xx_led.c
 *
 * Description: 
 *
 *    This file contains bcm963xx board led control API functions. 
 *
 *    To use it, do the following
 *
 *    1). define in the board.c the following led mappping (this is for 6345GW board):
 *        const LED_MAP_PAIR cLedMapping45GW[] =
 *        {   // led name     Initial state       physical pin (ledMask)
 *            {kLedUsb,       kLedStateOff,       GPIO_LED_PIN_7}, 
 *            {kLedAdsl,      kLedStateOff,       GPIO_LED_PIN_8},
 *            {kLedPPP,       kLedStateOff,       GPIO_LED_PIN_9},    // PPP and WanData share PIN_9
 *            {kLedWanData,   kLedStateOff,       GPIO_LED_PIN_9},
 *            {kLedWireless,  kLedStateOff,       GPIO_LED_PIN_10},
 *            {kLedEnd,       kLedStateOff,       0              } // NOTE: kLedEnd has to be at the end.
 *
 *    2). };To initialize led API and initial state of the leds, call the following function with the mapping 
 *        pointer from the above struct
 *
 *        boardLedInit((PLED_MAP_PAIR) &cLedMapping45R);
 *
 *    3). Sample call for kernel mode:
 *
 *        kerSysLedCtrl(kLedAdsl, kLedStateBlinkOnce);        // kLedxxx defines in board.h
 *
 *    4). Sample call for user mode
 *
 *        sysLedCtrl(kLedAdsl, kLedStateBlinkOnce);           // kLedxxx defines in board_api.h
 *
 *
 * Created on :  10/28/2002  seanl
 *
 ***************************************************************************/

/* Includes. */
#include <linux/fs.h>
#include <linux/iobuf.h>
#include <linux/capability.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/module.h>
#include <asm/uaccess.h>

#include <6345_map_part.h>
#include <board.h>

#define k100ms              ((100 * HZ) / 1000)     // ~100 ms
#define kFastBlinkCount     1                       // ~200ms
#define kSlowBlinkCount     5                       // ~600ms

// uncomment // for debug led
//#define DEBUG_LED

// global variables:
struct timer_list gLedTimer;
int gTimerOn = FALSE;
int gLedCount = 0;
PLED_MAP_PAIR gLedMapPtr;

typedef struct ledinfo
{
    unsigned short ledMask;         // mask for led: ie. giop 10 = 0x0400
    BOARD_LED_STATE ledState;       // current led state
    BOARD_LED_STATE savedLedState;  // used in blink once for restore to the orignal ledState
    int blinkCountDown;             // if == 0, do blink (toggle).  Is assgined value and dec by 1 at each timer.
} LED_INFO, *PLED_INFO;

PLED_INFO gLed = NULL;              // content sample: {0x0100, kLedStateOff, 0, 0},

void ledTimerExpire(void);

//**************************************************************************************
// LED operations
//**************************************************************************************

// return the led info ptr associated with ledName
PLED_INFO ledMapping(BOARD_LED_NAME ledName)
{
    PLED_MAP_PAIR pLedMap = gLedMapPtr;
    int i;

    while (pLedMap->ledName != kLedEnd)
    {

        if (pLedMap->ledName == ledName)
            break;
        pLedMap++;
    }

    for (i = 0; i < gLedCount; i++)
    {
        if ((gLed+i)->ledMask == pLedMap->ledMask)
            return (gLed+i);
    }

    return (PLED_INFO) NULL;        // should not come here
}


// turn led on and set the ledState
void ledOn(PLED_INFO pLed)
{
    GPIO->GPIODir |= pLed->ledMask;         // turn on the direction bit in case was turned off by some one
    GPIO->GPIOio  |= pLed->ledMask;         // turn on the led
    pLed->ledState = pLed->savedLedState = kLedStateOn;
}


// turn led off and set the ledState
void ledOff(PLED_INFO pLed)
{
    GPIO->GPIODir |= pLed->ledMask;         // turn on the direction bit in case was turned off by some one
    GPIO->GPIOio  &= ~(pLed->ledMask);      // turn off the led
    pLed->ledState = pLed->savedLedState = kLedStateOff;
}

 
// toggle the led and return the current ledState
BOARD_LED_STATE ledToggle(PLED_INFO pLed)
{
    GPIO->GPIODir |= pLed->ledMask;         // turn on the direction bit in case was turned off by some one
    if (GPIO->GPIOio & pLed->ledMask)       // it is on, turn it off
    {
        GPIO->GPIOio &= ~(pLed->ledMask);
        return kLedStateOff;
    }
    else
    {
        GPIO->GPIOio |= pLed->ledMask;      // it is off, turn it on
        return kLedStateOn;
    }
}   


// led timer.  Will return if timer is already on
void ledTimerStart(void)
{
    if (gTimerOn)
        return;

#if defined(DEBUG_LED)
    printk("led: add_timer\n");
#endif

    init_timer(&gLedTimer);
    gLedTimer.function = (void*)ledTimerExpire;
    gLedTimer.expires = jiffies + k100ms;        // timer expires in ~100ms
    add_timer (&gLedTimer);
    gTimerOn = TRUE;
} 


// led timer expire kicks in about ~100ms and perform the led operation according to the ledState and
// restart the timer according to ledState
void ledTimerExpire(void)
{
    int i;
    PLED_INFO pCurLed;

    gTimerOn = FALSE;

    for (i = 0, pCurLed = gLed; i < gLedCount; i++, pCurLed++)
    {
#if defined(DEBUG_LED)
        printk("led[%d]: Mask=0x%04x, State = %d, blcd=%d\n", i, pCurLed->ledMask, pCurLed->ledState, pCurLed->blinkCountDown);
#endif
        switch (pCurLed->ledState)
        {
            case kLedStateOn:
            case kLedStateOff:
                pCurLed->blinkCountDown = 0;            // reset the blink count down
                break;

            case kLedStateBlinkOnce:
                ledToggle(pCurLed);
                pCurLed->blinkCountDown = 0;                      // reset to 0
                pCurLed->ledState = pCurLed->savedLedState;
                if (pCurLed->ledState == kLedStateSlowBlinkContinues || 
                    pCurLed->ledState == kLedStateFastBlinkContinues)
                    ledTimerStart();                  // start timer if in blinkContinues stats
                break;

            case kLedStateSlowBlinkContinues:
                if (pCurLed->blinkCountDown-- == 0)
                {
                    pCurLed->blinkCountDown = kSlowBlinkCount;
                    ledToggle(pCurLed);
                }
                ledTimerStart();
                break;

            case kLedStateFastBlinkContinues:
                if (pCurLed->blinkCountDown-- == 0)
                {
                    pCurLed->blinkCountDown = kFastBlinkCount;
                    ledToggle(pCurLed);
                }
                ledTimerStart();
                break;

            default:
                printk("Invalid state = %d\n", pCurLed->ledState);
        }
    }
}

// initialize the gLedCount and allocate and fill gLed struct
void __init boardLedInit(PLED_MAP_PAIR cLedMapping)
{
    PLED_MAP_PAIR pLedMap = gLedMapPtr = cLedMapping;
    PLED_INFO pCurLed;
    int needTimer = FALSE;

    gLedCount = 0;

    // check for multiple ledName shares the same GPIO_LED
    while (pLedMap->ledName != kLedEnd)
    {
        int mask = pLedMap->ledMask;

        pLedMap++;
        if (pLedMap->ledName == kLedEnd)
        {
            gLedCount++;
            break;
        }
        if (pLedMap->ledMask != mask)
            gLedCount++;
    }

    gLed = (PLED_INFO) kmalloc((gLedCount * sizeof(LED_INFO)), GFP_KERNEL);

    // initial the gLed with unique ledMask and initial state. If more than 1 ledNames share the physical led 
    // (ledMask) the first defined led's ledInitState will be used.
    pLedMap = gLedMapPtr;
    pCurLed = gLed;
    do 
    {
        PLED_MAP_PAIR pCurMap;

        pCurMap = pLedMap++;
        if (pLedMap->ledName == kLedEnd || pCurMap->ledMask != pLedMap->ledMask)       // insert the Led info to gLed
        {
            pCurLed->ledState = pCurLed->savedLedState = pCurMap->ledInitState;
            pCurLed->ledMask = pCurMap->ledMask;
            switch (pCurLed->ledState)
            {
                case kLedStateOn:
                    pCurLed->blinkCountDown = 0;            // reset the blink count down
                    ledOn(pCurLed);
                    break;
                case kLedStateOff:
                    pCurLed->blinkCountDown = 0;            // reset the blink count down
                    ledOff(pCurLed);
                    break;
                case kLedStateBlinkOnce:
                    pCurLed->blinkCountDown = 1;
                    needTimer = TRUE;
                    break;
                case kLedStateSlowBlinkContinues:
                    pCurLed->blinkCountDown = kSlowBlinkCount;
                    needTimer = TRUE;
                    break;
                case kLedStateFastBlinkContinues:
                    pCurLed->blinkCountDown = kFastBlinkCount;
                    needTimer = TRUE;
                    break;
                default:
                    printk("Invalid state = %d\n", pCurLed->ledState);
            }
            pCurLed++;
        }
    } while(pLedMap->ledName != kLedEnd);


    if (needTimer)
        ledTimerStart();

#if defined(DEBUG_LED)
    int i;
    for (i=0; i < gLedCount; i++)
        printk("initLed: led[%d]: mask=0x%04x, state=%d\n", i,(gLed+i)->ledMask, (gLed+i)->ledState);
#endif

}

// led ctrl.  Maps the ledName to the corresponding ledInfoPtr and perform the led operation
void boardLedCtrl(BOARD_LED_NAME ledName, BOARD_LED_STATE ledState)
{
    PLED_INFO ledInfoPtr;

    // do the mapping from virtual to physical led
    if ((ledInfoPtr = ledMapping(ledName)) == NULL)
        return;

#if 1 /* Change for USR */
    // Do not blink the WAN Data LED if at least one interface is in bridge mode.
    if( ledName == kLedWanData )
    {
        static int isBridgedProtocol(void);
        static int BridgedProtocol = -1;

        if( BridgedProtocol == -1 )
            BridgedProtocol = isBridgedProtocol();

        if( BridgedProtocol == TRUE )
            return;
    }
#endif

    switch (ledState)
    {
        case kLedStateOn:
            ledOn(ledInfoPtr);
            break;

        case kLedStateOff: 
            ledOff(ledInfoPtr);
            break;

        case kLedStateBlinkOnce:
            // skip blinkOnce if it is already in Slow/Fast blink continues state
//            if (ledInfoPtr->savedLedState == kLedStateSlowBlinkContinues ||
//                ledInfoPtr->savedLedState == kLedStateFastBlinkContinues)
//                ;
//            else
            if( ledInfoPtr->savedLedState == kLedStateOn )  //USR9106: blink led only if it is On.
            {
                if (ledInfoPtr->blinkCountDown == 0)  // skip the call if it is 1
                {
                    ledToggle(ledInfoPtr);
                    ledInfoPtr->blinkCountDown = 1;  // it will be reset to 0 when timer expires
                    ledInfoPtr->ledState = kLedStateBlinkOnce;
                    ledTimerStart();
                }
            }
            break;

        case kLedStateSlowBlinkContinues:
            ledInfoPtr->blinkCountDown = kSlowBlinkCount;
            ledInfoPtr->ledState = kLedStateSlowBlinkContinues;
            ledInfoPtr->savedLedState = kLedStateSlowBlinkContinues;
            ledTimerStart();
            break;

        case kLedStateFastBlinkContinues:
            ledInfoPtr->blinkCountDown = kFastBlinkCount;
            ledInfoPtr->ledState = kLedStateFastBlinkContinues;
            ledInfoPtr->savedLedState = kLedStateFastBlinkContinues;
            ledTimerStart();
            break;

        default:
            printk("Invalid led state\n");
    }
}

#if 1 /* Change for USR */
// Determines if there is at least one interface in bridge mode.  Bridge mode
// is determined by the cfm convention of naming bridge interfaces nas17
// through nas24.
static int isBridgedProtocol(void)
{
    extern int dev_get(const char *name);
    const int firstBridgeId = 17;
    const int lastBridgeId = 24;
    int i;
    int ret = FALSE;
    char name[16];

    for( i = firstBridgeId; i <= lastBridgeId; i++ )
    {
        sprintf( name, "nas%d", i );

        if( dev_get(name) )
        {
            ret = TRUE;
            break;
        }
    }

    return(ret);
}
#endif

